from dataclasses import dataclass, field
from typing import Any


@dataclass
class CollectionItem:
    name: str
    slug: str
    is_archived: bool = False
    is_draft: bool = False
    fields: dict[str, Any] = field(default_factory=dict)
    id: str | None = None


@dataclass
class CollectionField:
    slug: str
    is_required: bool
    is_editable: bool



from typing import Any, Sequence
from urllib.parse import urljoin

import requests

from .types import CollectionItem, CollectionField

DEFAULT_BASE_URL = "https://api.webflow.com/v2/"


class WebflowClient:
    def __init__(self, access_token: str, base_url: str | None = None):
        self._access_token = access_token
        self._base_url = base_url if base_url is not None else DEFAULT_BASE_URL

    def _request(
        self, method: str, path: str, *, params: dict[str, Any] | None = None, data=None
    ):
        response = requests.request(
            method,
            urljoin(self._base_url, path),
            headers={
                "Authorization": f"Bearer {self._access_token}",
            },
            params=params,
            json=data,
        )

        if not response.ok:
            raise ValueError(
                f"Unable to {method} {path}: {response.status_code} {response.text}"
            )

        return response.json()

    @staticmethod
    def _parse_collection_item(data) -> CollectionItem:
        fields = data.get("fieldData") or {}

        name = fields.pop("name")
        slug = fields.pop("slug")

        return CollectionItem(
            id=data["id"],
            name=name,
            slug=slug,
            is_archived=data.get("isArchived", False),
            is_draft=data.get("isDraft", False),
            fields=fields,
        )

    @staticmethod
    def _serialize_collection_item(value: CollectionItem):
        return {
            "id": value.id,
            "isArchived": value.is_archived,
            "isDraft": value.is_draft,
            "fieldData": {
                **value.fields,
                "name": value.name,
                "slug": value.slug,
            },
        }

    def get_item_by_slug(self, collection_id: str, slug: str) -> CollectionItem | None:
        data = self._request(
            "GET", f"collections/{collection_id}/items", params={"slug": slug}
        )

        if len(data["items"]) == 0:
            return None

        return self._parse_collection_item(data["items"][0])

    @staticmethod
    def _parse_collection_field(data) -> CollectionField:
        return CollectionField(
            slug=data["slug"],
            is_required=data["isRequired"],
            is_editable=data["isEditable"],
        )

    def get_collection_schema(self, collection_id: str) -> Sequence[CollectionField]:
        collection = self._request("GET", f"collections/{collection_id}")

        return [self._parse_collection_field(f) for f in collection["fields"]]

    def insert_collection_items(
        self, collection_id: str, items: Sequence[CollectionItem]
    ) -> Sequence[CollectionItem]:
        response = self._request(
            "POST",
            f"collections/{collection_id}/items",
            data={
                "items": [self._serialize_collection_item(i) for i in items],
            },
        )

        return [self._parse_collection_item(i) for i in response["items"]]

    def update_collection_items(
        self, collection_id: str, items: Sequence[CollectionItem]
    ) -> Sequence[CollectionItem]:
        response = self._request(
            "PATCH",
            f"collections/{collection_id}/items",
            data={
                "items": [self._serialize_collection_item(i) for i in items],
            },
        )

        return [self._parse_collection_item(i) for i in response["items"]]

    def publish_collection_items(
        self, collection_id: str, items: Sequence[CollectionItem]
    ):
        self._request(
            "POST",
            f"collections/{collection_id}/items/publish",
            data={
                "itemIds": [i.id for i in items if i.id is not None],
            },
        )


__all__ = ("WebflowClient",)



import pytest
from pytest_mock import MockType, MockerFixture

from bytewax_webflow.outputs.collection_item_sink import (
    WebflowCollectionItemSinkPartition,
)
from bytewax_webflow.types import WebflowCollectionItem, WebflowCollectionField
from bytewax_webflow.webflow_client import WebflowClient


class TestWebflowCollectionItemSinkPartition:
    @pytest.fixture
    def mock_client(self, mocker: MockerFixture):
        mock = mocker.create_autospec(spec=WebflowClient)

        mock.get_item_by_slug.return_value = None

        mock.get_collection_schema.return_value = [
            WebflowCollectionField(
                slug="example",
                is_required=False,
                is_editable=True,
            ),
            WebflowCollectionField(
                slug="example-two",
                is_required=True,
                is_editable=True,
            ),
            WebflowCollectionField(
                slug="example-three",
                is_required=False,
                is_editable=False,
            ),
        ]

        yield mock

    def test_inserts_valid_records_no_slug(self, mock_client: MockType):
        partition = WebflowCollectionItemSinkPartition(
            mock_client, "example-collection"
        )

        input_item = WebflowCollectionItem(
            name="Example",
            slug="example",
            fields={
                "example": "Accepted field",
                "example-two": "Required Field",
            },
        )

        partition.write_batch([input_item])

        mock_client.insert_collection_items.assert_called_with(
            "example-collection", [input_item]
        )
        mock_client.update_collection_items.assert_not_called()

    def test_updates_records_with_id(self, mock_client: MockType):
        partition = WebflowCollectionItemSinkPartition(
            mock_client, "example-collection"
        )

        input_item = WebflowCollectionItem(
            id="example-id",
            name="Example",
            slug="example",
            fields={
                "example": "Accepted field",
                "example-two": "Required Field",
            },
        )

        partition.write_batch([input_item])

        mock_client.insert_collection_items.assert_not_called()
        mock_client.update_collection_items.assert_called_with(
            "example-collection", [input_item]
        )

    def test_updates_records_when_slug_overlaps(self, mock_client: MockType):
        partition = WebflowCollectionItemSinkPartition(
            mock_client, "example-collection"
        )

        from_slug_item = WebflowCollectionItem(
            id="example-id",
            name="Example",
            slug="example",
            fields={
                "example": "Accepted field",
                "example-two": "Required Field",
            },
        )

        input_item = WebflowCollectionItem(
            name="Example",
            slug="example",
            fields={
                "example": "Overridden data",
                "example-two": "Required Field",
            },
        )

        expected = WebflowCollectionItem(
            id="example-id",
            name="Example",
            slug="example",
            fields={
                "example": "Overridden data",
                "example-two": "Required Field",
            },
        )

        mock_client.get_item_by_slug.return_value = from_slug_item

        partition.write_batch([input_item])

        mock_client.insert_collection_items.assert_not_called()
        mock_client.update_collection_items.assert_called_with(
            "example-collection", [expected]
        )

    def test_ignores_records_with_invalid_keys(self, mock_client: MockType):
        partition = WebflowCollectionItemSinkPartition(
            mock_client, "example-collection"
        )

        input_item = WebflowCollectionItem(
            name="Example",
            slug="example",
            fields={
                "example": "Accepted field",
                "example-two": "Required Field",
                "invalid": "Invalid field",
            },
        )

        partition.write_batch([input_item])

        mock_client.insert_collection_items.assert_not_called()
        mock_client.update_collection_items.assert_not_called()

    def test_ignores_records_with_missing_required_keys(self, mock_client: MockType):
        partition = WebflowCollectionItemSinkPartition(
            mock_client, "example-collection"
        )

        input_item = WebflowCollectionItem(
            name="Example",
            slug="example",
            fields={
                "example": "Accepted field",
            },
        )

        partition.write_batch([input_item])

        mock_client.insert_collection_items.assert_not_called()
        mock_client.update_collection_items.assert_not_called()

    def test_ignores_records_with_uneditable_keys(self, mock_client: MockType):
        partition = WebflowCollectionItemSinkPartition(
            mock_client, "example-collection"
        )

        input_item = WebflowCollectionItem(
            name="Example",
            slug="example",
            fields={
                "example": "Accepted field",
                "example-two": "Required Field",
                "example-three": "Uneditable Field",
            },
        )

        partition.write_batch([input_item])

        mock_client.insert_collection_items.assert_not_called()
        mock_client.update_collection_items.assert_not_called()



from logging import getLogger
from typing import Sequence

from bytewax.outputs import (
    StatelessSinkPartition,
    DynamicSink,
)

from ..webflow_client import WebflowClient
from ..types import CollectionItem, CollectionField


logger = getLogger(__name__)


class CollectionItemSinkPartition(StatelessSinkPartition[CollectionItem]):
    def __init__(
        self,
        client: WebflowClient,
        collection_id: str,
        is_publishing: bool = False,
    ):
        self._client = client
        self._collection_id = collection_id
        self._is_publishing = is_publishing

    @staticmethod
    def _is_valid(schema: Sequence[CollectionField], value: CollectionItem):
        valid_fields = set(
            [f.slug for f in schema if f.is_editable and f.slug not in ["slug", "name"]]
        )
        required_fields = set(
            [f.slug for f in schema if f.is_required and f.slug not in ["slug", "name"]]
        )

        value_fields = set(value.fields.keys())

        if (
            not value.is_draft
            and len(required_fields) > 0
            and not required_fields <= value_fields
        ):
            logger.warning(
                f"Item with slug {value.slug} missing required fields: {", ".join(required_fields.difference(value_fields))}"
            )
            return False

        if not value_fields <= valid_fields:
            logger.warning(
                f"Item with slug {value.slug} has invalid fields: {", ".join(value_fields.difference(valid_fields))}"
            )
            return False

        return True

    def write_batch(self, items: Sequence[CollectionItem]) -> None:
        if len(items) == 0:
            return

        schema = self._client.get_collection_schema(self._collection_id)

        inserts = []
        updates = []

        for item in items:
            if not self._is_valid(schema, item):
                continue

            existing_item = self._client.get_item_by_slug(
                self._collection_id, item.slug
            )

            if existing_item is not None:
                existing_item.is_archived = item.is_archived
                existing_item.is_draft = item.is_draft
                existing_item.name = item.name

                # Update only set values
                for key, value in item.fields.items():
                    existing_item.fields[key] = value

                item = existing_item

            if item.id is not None:
                updates.append(item)
            else:
                inserts.append(item)

        touched_items = []

        if len(inserts) > 0:
            logger.info(f"Inserted {len(inserts)} collection items")

            touched_items += self._client.insert_collection_items(
                self._collection_id, inserts
            )

        if len(updates) > 0:
            logger.info(f"Updated {len(updates)} collection items")

            touched_items += self._client.update_collection_items(
                self._collection_id, updates
            )

        if self._is_publishing and len(touched_items) > 0:
            self._client.publish_collection_items(self._collection_id, touched_items)


class CollectionItemSink(DynamicSink[CollectionItem]):
    """Fixed partitioned sink for writing data to a Webflow Collection."""

    def __init__(
        self,
        access_token: str,
        collection_id: str,
        is_publishing: bool = False,
        base_url: str | None = None,
    ) -> None:
        """Initialize the WebflowCollectionSink."""
        self._access_token = access_token
        self._collection_id = collection_id
        self._is_publishing = is_publishing
        self._base_url = base_url

    def build(
        self,
        step_id: str,
        worker_index: int,
        worker_count: int,
    ) -> CollectionItemSinkPartition:
        """Build or resume a partition."""
        if worker_count != 1:
            raise ValueError("only supports 1 worker")

        client = WebflowClient(
            access_token=self._access_token,
            base_url=self._base_url,
        )

        return CollectionItemSinkPartition(
            client=client,
            collection_id=self._collection_id,
            is_publishing=self._is_publishing,
        )


__all__ = ("CollectionItemSinkPartition", "CollectionItemSink")
